package org.scalaquery.ql

import scala.annotation.implicitNotFound
import org.scalaquery.util.{Node, ValueLinearizer, ProductLinearizer, WithOp}

/**
 * A type class that encodes the unpacking `Packed => Unpacked` of a
 * `Query[Packed]` to its result element type `Unpacked`.
 *
 * At run-time, it provides the reification, `ValueLinearizer` of the
 * `Unpacked` type.
 *
 * =Example:=
 * - Packed: (Column[Int], Column[(Int, String)], (Int, Option[Double]))
 * - Unpacked: (Int, (Int, String), (Int, Option[Double]))
 * - Reified: (Column[Int], Column[(Int, String)], (Column[Int], Column[Option[Double]]))
 * - Linearized: (Int, Int, String, Int, Option[Double])
 */
@implicitNotFound(msg = "Don't know how to unpack ${Packed_} (to ${Unpacked_})")
sealed trait Unpack[-Packed_, Unpacked_] {
  type Packed = Packed_
  type Unpacked = Unpacked_
  def reify(from: Packed): Any = from // must return the reified type
  def reifiedUnpack: Unpack[_, Unpacked] = this // must return an Unpack for the reified type
  def linearizer(from: Packed): ValueLinearizer[_]
  def mapOp(x: Packed, f: Node => Node): Any // really returns x.type
}

object Unpack extends UnpackLowPriority {
  def unpackValueLinearizer[T <: ValueLinearizer[_] with WithOp, U] = new Unpack[T, U] {
    def linearizer(from: Packed): ValueLinearizer[_] = from
    def mapOp(from: Packed, f: Node => Node) = from.mapOp(f)
  }

  implicit def unpackJoin[T1 <: AbstractTable[_], T2 <: AbstractTable[_]]: Unpack[Join[T1, T2], (T1, T2)] = new Unpack[Join[T1, T2], (T1, T2)] {
    def linearizer(from: Packed) = from.left
    def mapOp(from: Packed, f: Node => Node) = from.mapOp(f)
  }

  def unpackTableBase[T <: TableBase[_]]: Unpack[T, Nothing] = new Unpack[T, Nothing] {
    def linearizer(from: Packed): ValueLinearizer[_] = from match {
      case a: AbstractTable[_] => a.*
      case j: Join[_, _] => j.left.*
    }
    def mapOp(from: Packed, f: Node => Node) = from.mapOp(f)
  }
}

trait UnpackLowPriority extends UnpackLowPriority2 {
  implicit final def unpackColumnBase[T]: Unpack[ColumnBase[T], T] = new Unpack[ColumnBase[T], T] {
    def linearizer(from: Packed) = from
    def mapOp(from: Packed, f: Node => Node) = from.mapOp(f)
  }

  implicit final def unpackPrimitive[T](implicit tm: TypeMapper[T]): Unpack[T, T] = new Unpack[T, T] {
    override def reify(from: Packed) = ConstColumn(from)
    override def reifiedUnpack = unpackColumnBase[T]
    def linearizer(from: Packed) = ConstColumn(from)
    def mapOp(from: Packed, f: Node => Node) = from
  }
}

trait UnpackLowPriority2 {
  implicit final def unpackTuple2[T1,T2, U1,U2, R1,R2](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2]): Unpack[(T1,T2), (U1,U2)] = new Unpack[(T1,T2), (U1,U2)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple2(u1.reifiedUnpack, u2.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f))
  }
  implicit final def unpackTuple3[T1,T2,T3, U1,U2,U3, R1,R2,R3](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3]): Unpack[(T1,T2,T3), (U1,U2,U3)] = new Unpack[(T1,T2,T3), (U1,U2,U3)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple3(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f))
  }
  implicit final def unpackTuple4[T1,T2,T3,T4, U1,U2,U3,U4, R1,R2,R3,R4](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4]): Unpack[(T1,T2,T3,T4), (U1,U2,U3,U4)] = new Unpack[(T1,T2,T3,T4), (U1,U2,U3,U4)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple4(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f))
  }
  implicit final def unpackTuple5[T1,T2,T3,T4,T5, U1,U2,U3,U4,U5, R1,R2,R3,R4,R5](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5]): Unpack[(T1,T2,T3,T4,T5), (U1,U2,U3,U4,U5)] = new Unpack[(T1,T2,T3,T4,T5), (U1,U2,U3,U4,U5)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple5(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f))
  }
  implicit final def unpackTuple6[T1,T2,T3,T4,T5,T6, U1,U2,U3,U4,U5,U6, R1,R2,R3,R4,R5,R6](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6]): Unpack[(T1,T2,T3,T4,T5,T6), (U1,U2,U3,U4,U5,U6)] = new Unpack[(T1,T2,T3,T4,T5,T6), (U1,U2,U3,U4,U5,U6)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple6(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f))
  }
  implicit final def unpackTuple7[T1,T2,T3,T4,T5,T6,T7, U1,U2,U3,U4,U5,U6,U7, R1,R2,R3,R4,R5,R6,R7](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7]): Unpack[(T1,T2,T3,T4,T5,T6,T7), (U1,U2,U3,U4,U5,U6,U7)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7), (U1,U2,U3,U4,U5,U6,U7)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple7(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f))
  }
  implicit final def unpackTuple8[T1,T2,T3,T4,T5,T6,T7,T8, U1,U2,U3,U4,U5,U6,U7,U8, R1,R2,R3,R4,R5,R6,R7,R8](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8), (U1,U2,U3,U4,U5,U6,U7,U8)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8), (U1,U2,U3,U4,U5,U6,U7,U8)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple8(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f))
  }
  implicit final def unpackTuple9[T1,T2,T3,T4,T5,T6,T7,T8,T9, U1,U2,U3,U4,U5,U6,U7,U8,U9, R1,R2,R3,R4,R5,R6,R7,R8,R9](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9), (U1,U2,U3,U4,U5,U6,U7,U8,U9)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9), (U1,U2,U3,U4,U5,U6,U7,U8,U9)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple9(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f))
  }
  implicit final def unpackTuple10[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple10(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f))
  }
  implicit final def unpackTuple11[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple11(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f))
  }
  implicit final def unpackTuple12[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple12(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f))
  }
  implicit final def unpackTuple13[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple13(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f))
  }
  implicit final def unpackTuple14[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple14(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f))
  }
  implicit final def unpackTuple15[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple15(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f))
  }
  implicit final def unpackTuple16[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple16(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f))
  }
  implicit final def unpackTuple17[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple17(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f))
  }
  implicit final def unpackTuple18[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17], u18: Unpack[T18, U18]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17), u18.reify(from._18))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple18(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack, u18.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17), u18.linearizer(from._18))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f), u18.mapOp(from._18, f))
  }
  implicit final def unpackTuple19[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17], u18: Unpack[T18, U18], u19: Unpack[T19, U19]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17), u18.reify(from._18), u19.reify(from._19))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple19(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack, u18.reifiedUnpack, u19.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17), u18.linearizer(from._18), u19.linearizer(from._19))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f), u18.mapOp(from._18, f), u19.mapOp(from._19, f))
  }
  implicit final def unpackTuple20[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17], u18: Unpack[T18, U18], u19: Unpack[T19, U19], u20: Unpack[T20, U20]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17), u18.reify(from._18), u19.reify(from._19), u20.reify(from._20))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple20(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack, u18.reifiedUnpack, u19.reifiedUnpack, u20.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17), u18.linearizer(from._18), u19.linearizer(from._19), u20.linearizer(from._20))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f), u18.mapOp(from._18, f), u19.mapOp(from._19, f), u20.mapOp(from._20, f))
  }
  implicit final def unpackTuple21[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17], u18: Unpack[T18, U18], u19: Unpack[T19, U19], u20: Unpack[T20, U20], u21: Unpack[T21, U21]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17), u18.reify(from._18), u19.reify(from._19), u20.reify(from._20), u21.reify(from._21))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple21(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack, u18.reifiedUnpack, u19.reifiedUnpack, u20.reifiedUnpack, u21.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17), u18.linearizer(from._18), u19.linearizer(from._19), u20.linearizer(from._20), u21.linearizer(from._21))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f), u18.mapOp(from._18, f), u19.mapOp(from._19, f), u20.mapOp(from._20, f), u21.mapOp(from._21, f))
  }
  implicit final def unpackTuple22[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21,U22, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21,R22](implicit u1: Unpack[T1, U1], u2: Unpack[T2, U2], u3: Unpack[T3, U3], u4: Unpack[T4, U4], u5: Unpack[T5, U5], u6: Unpack[T6, U6], u7: Unpack[T7, U7], u8: Unpack[T8, U8], u9: Unpack[T9, U9], u10: Unpack[T10, U10], u11: Unpack[T11, U11], u12: Unpack[T12, U12], u13: Unpack[T13, U13], u14: Unpack[T14, U14], u15: Unpack[T15, U15], u16: Unpack[T16, U16], u17: Unpack[T17, U17], u18: Unpack[T18, U18], u19: Unpack[T19, U19], u20: Unpack[T20, U20], u21: Unpack[T21, U21], u22: Unpack[T22, U22]): Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21,U22)] = new Unpack[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22), (U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21,U22)] {
    override def reify(from: Packed) =
      (u1.reify(from._1), u2.reify(from._2), u3.reify(from._3), u4.reify(from._4), u5.reify(from._5), u6.reify(from._6), u7.reify(from._7), u8.reify(from._8), u9.reify(from._9), u10.reify(from._10), u11.reify(from._11), u12.reify(from._12), u13.reify(from._13), u14.reify(from._14), u15.reify(from._15), u16.reify(from._16), u17.reify(from._17), u18.reify(from._18), u19.reify(from._19), u20.reify(from._20), u21.reify(from._21), u22.reify(from._22))
    override def reifiedUnpack: Unpack[_, Unpacked] =
      unpackTuple22(u1.reifiedUnpack, u2.reifiedUnpack, u3.reifiedUnpack, u4.reifiedUnpack, u5.reifiedUnpack, u6.reifiedUnpack, u7.reifiedUnpack, u8.reifiedUnpack, u9.reifiedUnpack, u10.reifiedUnpack, u11.reifiedUnpack, u12.reifiedUnpack, u13.reifiedUnpack, u14.reifiedUnpack, u15.reifiedUnpack, u16.reifiedUnpack, u17.reifiedUnpack, u18.reifiedUnpack, u19.reifiedUnpack, u20.reifiedUnpack, u21.reifiedUnpack, u22.reifiedUnpack)
    def linearizer(from: Packed) = {
      val sub = Vector(u1.linearizer(from._1), u2.linearizer(from._2), u3.linearizer(from._3), u4.linearizer(from._4), u5.linearizer(from._5), u6.linearizer(from._6), u7.linearizer(from._7), u8.linearizer(from._8), u9.linearizer(from._9), u10.linearizer(from._10), u11.linearizer(from._11), u12.linearizer(from._12), u13.linearizer(from._13), u14.linearizer(from._14), u15.linearizer(from._15), u16.linearizer(from._16), u17.linearizer(from._17), u18.linearizer(from._18), u19.linearizer(from._19), u20.linearizer(from._20), u21.linearizer(from._21), u22.linearizer(from._22))
      new ProductLinearizer(sub)
    }
    def mapOp(from: Packed, f: Node => Node) =
      (u1.mapOp(from._1, f), u2.mapOp(from._2, f), u3.mapOp(from._3, f), u4.mapOp(from._4, f), u5.mapOp(from._5, f), u6.mapOp(from._6, f), u7.mapOp(from._7, f), u8.mapOp(from._8, f), u9.mapOp(from._9, f), u10.mapOp(from._10, f), u11.mapOp(from._11, f), u12.mapOp(from._12, f), u13.mapOp(from._13, f), u14.mapOp(from._14, f), u15.mapOp(from._15, f), u16.mapOp(from._16, f), u17.mapOp(from._17, f), u18.mapOp(from._18, f), u19.mapOp(from._19, f), u20.mapOp(from._20, f), u21.mapOp(from._21, f), u22.mapOp(from._22, f))
  }
}

/**
 * A type class that encodes the reification `Packed => Reified` of a
 * `Query[Packed]` to a fully packed type `Reified`, i.e. a type where
 * everything which is not a transparent container is wrapped in a `Column[_]`.
 *
 * This needs to be defined in a separate typeclass but in lockstep with Unpack
 * because of limitations of Scala's type inferencer.
 *
 * =Example:=
 * - Packed: (Column[Int], Column[(Int, String)], (Int, Option[Double]))
 * - Reified: (Column[Int], Column[(Int, String)], (Column[Int], Column[Option[Double]]))
 * - Reified: (Column[Int], Column[(Int, String)], (Column[Int], Column[Option[Double]]))
 */
@implicitNotFound(msg = "Don't know how to reify ${Packed} (to ${Reified})")
sealed trait Reify[-Packed, Reified]

object Reify extends ReifyLowPriority {
  implicit final def reifyColumnBase[T <: ColumnBase[_]]: Reify[T, T] = null
}

trait ReifyLowPriority {
  implicit final def reifyPrimitive[T](implicit tm: TypeMapper[T]): Reify[T, ConstColumn[T]] = null

  implicit final def reifyTuple2[T1,T2, U1,U2, R1,R2](implicit u1: Reify[T1, R1], u2: Reify[T2, R2]): Reify[(T1,T2), (R1,R2)] = null
  implicit final def reifyTuple3[T1,T2,T3, U1,U2,U3, R1,R2,R3](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3]): Reify[(T1,T2,T3), (R1,R2,R3)] = null
  implicit final def reifyTuple4[T1,T2,T3,T4, U1,U2,U3,U4, R1,R2,R3,R4](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4]): Reify[(T1,T2,T3,T4), (R1,R2,R3,R4)] = null
  implicit final def reifyTuple5[T1,T2,T3,T4,T5, U1,U2,U3,U4,U5, R1,R2,R3,R4,R5](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5]): Reify[(T1,T2,T3,T4,T5), (R1,R2,R3,R4,R5)] = null
  implicit final def reifyTuple6[T1,T2,T3,T4,T5,T6, U1,U2,U3,U4,U5,U6, R1,R2,R3,R4,R5,R6](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6]): Reify[(T1,T2,T3,T4,T5,T6), (R1,R2,R3,R4,R5,R6)] = null
  implicit final def reifyTuple7[T1,T2,T3,T4,T5,T6,T7, U1,U2,U3,U4,U5,U6,U7, R1,R2,R3,R4,R5,R6,R7](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7]): Reify[(T1,T2,T3,T4,T5,T6,T7), (R1,R2,R3,R4,R5,R6,R7)] = null
  implicit final def reifyTuple8[T1,T2,T3,T4,T5,T6,T7,T8, U1,U2,U3,U4,U5,U6,U7,U8, R1,R2,R3,R4,R5,R6,R7,R8](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8), (R1,R2,R3,R4,R5,R6,R7,R8)] = null
  implicit final def reifyTuple9[T1,T2,T3,T4,T5,T6,T7,T8,T9, U1,U2,U3,U4,U5,U6,U7,U8,U9, R1,R2,R3,R4,R5,R6,R7,R8,R9](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9), (R1,R2,R3,R4,R5,R6,R7,R8,R9)] = null
  implicit final def reifyTuple10[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10)] = null
  implicit final def reifyTuple11[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11)] = null
  implicit final def reifyTuple12[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12)] = null
  implicit final def reifyTuple13[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13)] = null
  implicit final def reifyTuple14[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14)] = null
  implicit final def reifyTuple15[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15)] = null
  implicit final def reifyTuple16[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16)] = null
  implicit final def reifyTuple17[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17)] = null
  implicit final def reifyTuple18[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17], u18: Reify[T18, R18]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18)] = null
  implicit final def reifyTuple19[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17], u18: Reify[T18, R18], u19: Reify[T19, R19]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19)] = null
  implicit final def reifyTuple20[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17], u18: Reify[T18, R18], u19: Reify[T19, R19], u20: Reify[T20, R20]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20)] = null
  implicit final def reifyTuple21[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17], u18: Reify[T18, R18], u19: Reify[T19, R19], u20: Reify[T20, R20], u21: Reify[T21, R21]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21)] = null
  implicit final def reifyTuple22[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22, U1,U2,U3,U4,U5,U6,U7,U8,U9,U10,U11,U12,U13,U14,U15,U16,U17,U18,U19,U20,U21,U22, R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21,R22](implicit u1: Reify[T1, R1], u2: Reify[T2, R2], u3: Reify[T3, R3], u4: Reify[T4, R4], u5: Reify[T5, R5], u6: Reify[T6, R6], u7: Reify[T7, R7], u8: Reify[T8, R8], u9: Reify[T9, R9], u10: Reify[T10, R10], u11: Reify[T11, R11], u12: Reify[T12, R12], u13: Reify[T13, R13], u14: Reify[T14, R14], u15: Reify[T15, R15], u16: Reify[T16, R16], u17: Reify[T17, R17], u18: Reify[T18, R18], u19: Reify[T19, R19], u20: Reify[T20, R20], u21: Reify[T21, R21], u22: Reify[T22, R22]): Reify[(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22), (R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14,R15,R16,R17,R18,R19,R20,R21,R22)] = null
}
